@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2021 taylor.fish <contact@taylor.fish>
+ * Copyright (C) 2021-2022 taylor.fish <contact@taylor.fish>
  *
  * This file is part of Base116.
  *
@@ -24,7 +24,7 @@ use super::{BYTES_PER_CHUNK, DIGITS_PER_CHUNK, END_CHAR, START_CHAR};
 use super::{L1_MULT, L2_MULT};
 
 use core::array;
-use core::convert::{Infallible, TryFrom};
+use core::convert::TryFrom;
 use core::fmt::{self, Debug, Display, Formatter};
 use core::iter::{FusedIterator, Take};
 use core::str::Chars;
@@ -42,14 +42,13 @@ pub enum DecodeError {
     TrailingData(char),
 }
 
-pub type DecodeResult<T> = Result<T, DecodeError>;
-pub type DecodeBytesResult<T> = Result<T, DecodeBytesError>;
 use DecodeError as Error;
+pub type DecodeResult<T> = Result<T, DecodeError>;
 
 impl fmt::Display for DecodeError {
     fn fmt(&self, f: &mut Formatter<'_>) -> Result<(), fmt::Error> {
         match self {
-            Self::BadChar(c) => write!(f, "bad character: {}", c),
+            Self::BadChar(c) => write!(f, "bad character: {:?}", c),
             Self::BadLength => write!(f, "bad input length"),
             Self::MissingStart => {
                 write!(f, "missing start character ({})", START_CHAR)
@@ -101,6 +100,8 @@ pub enum DecodeBytesError {
     DecodeError(DecodeError),
 }
 
+pub type DecodeBytesResult<T> = Result<T, DecodeBytesError>;
+
 impl fmt::Display for DecodeBytesError {
     fn fmt(&self, f: &mut Formatter<'_>) -> Result<(), fmt::Error> {
         match self {
@@ -252,7 +253,7 @@ where
             } else {
                 panic!(
                     "char ({:?}) unmapped to bad value: invalid digit: {}",
-                    c, n
+                    c, n,
                 );
             }
         }
@@ -349,7 +350,7 @@ where
                     len += 1;
                 })
             })
-            .and_then(|_| match len {
+            .and(match len {
                 0 => Ok(None),
                 1 => Err(Error::BadLength),
                 _ => Ok(Some(())),
@@ -383,11 +384,11 @@ impl<I: FusedIterator<Item = DecodeResult<Digit>>> FusedIterator
 }
 
 #[allow(clippy::type_complexity)]
-struct BaseDecoder<I>(
+pub struct CharDecoder<I>(
     Flatten<
         DigitsToUnflatBytes<
             Flatten<
-                CharsToUnflatDigits<UnwrapString<I>>,
+                CharsToUnflatDigits<UnwrapString<BaseIterator<I>>>,
                 CharsToUnflatDigitsItem,
             >,
         >,
@@ -395,50 +396,25 @@ struct BaseDecoder<I>(
     >,
 );
 
-impl<I> BaseDecoder<I> {
-    pub fn new(iter: I, config: DecodeConfig) -> Self {
+impl<I> CharDecoder<I> {
+    pub(crate) fn new(iter: I, config: DecodeConfig) -> Self {
         Self(Flatten::new(DigitsToUnflatBytes::new(Flatten::new(
-            CharsToUnflatDigits::new(UnwrapString::new(iter, config)),
+            CharsToUnflatDigits::new(UnwrapString::new(
+                BaseIterator(iter),
+                config,
+            )),
         ))))
     }
 }
 
-impl<I: InspectBaseIterator> InspectBaseIterator for BaseDecoder<I> {
-    type Iter = I::Iter;
+impl<I> InspectBaseIterator for CharDecoder<I> {
+    type Iter = I;
 
     fn base_iterator(&self) -> &Self::Iter {
         self.0.base_iterator()
     }
 }
 
-impl<I> Iterator for BaseDecoder<I>
-where
-    I: Iterator<Item = char>,
-{
-    type Item = DecodeResult<u8>;
-
-    fn next(&mut self) -> Option<Self::Item> {
-        self.0.next()
-    }
-
-    fn fold<B, F>(self, init: B, f: F) -> B
-    where
-        F: FnMut(B, Self::Item) -> B,
-    {
-        self.0.fold(init, f)
-    }
-}
-
-impl<I: FusedIterator<Item = char>> FusedIterator for BaseDecoder<I> {}
-
-pub struct CharDecoder<I>(BaseDecoder<BaseIterator<I>>);
-
-impl<I> CharDecoder<I> {
-    pub(crate) fn new(iter: I, config: DecodeConfig) -> Self {
-        Self(BaseDecoder::new(BaseIterator(iter), config))
-    }
-}
-
 impl<I> Iterator for CharDecoder<I>
 where
     I: Iterator<Item = char>,
@@ -481,7 +457,9 @@ where
     }
 }
 
-impl<I: FusedIterator<Item = char>> FusedIterator for CharDecoder<I> {}
+// `UnwrapString` implements `FusedIterator`, so `CharDecoder` also can even
+// if `I` doesn't.
+impl<I: Iterator<Item = char>> FusedIterator for CharDecoder<I> {}
 
 fn size_hint_from_utf8_hint(
     hint: (usize, Option<usize>),
@@ -553,17 +531,13 @@ where
 impl<I: FusedIterator<Item = u8>> FusedIterator for Utf8ToChars<I> {}
 
 pub struct BytesDecoder<I>(
-    BaseDecoder<
-        BaseIterator<ErrAdapter<Utf8ToChars<BaseIterator<I>>, InvalidUtf8>>,
-    >,
+    CharDecoder<ErrAdapter<Utf8ToChars<BaseIterator<I>>, InvalidUtf8>>,
 );
 
 impl<I> BytesDecoder<I> {
     pub(crate) fn new(iter: I, config: DecodeConfig) -> Self {
-        Self(BaseDecoder::new(
-            BaseIterator(ErrAdapter::new(Utf8ToChars::new(BaseIterator(
-                iter,
-            )))),
+        Self(CharDecoder::new(
+            ErrAdapter::new(Utf8ToChars::new(BaseIterator(iter))),
             config,
         ))
     }
@@ -577,28 +551,46 @@ where
 
     fn next(&mut self) -> Option<Self::Item> {
         let item = self.0.next();
-        if let Some(e) = self.0.base_iterator().take_err() {
-            return Some(Err(DecodeBytesError::InvalidUtf8(e)));
-        }
-        match item {
-            Some(Ok(b)) => Some(Ok(b)),
-            Some(Err(e)) => Some(Err(DecodeBytesError::DecodeError(e))),
-            None => None,
+        if !matches!(item, Some(Ok(_))) {
+            if let Some(e) = self.0.base_iterator().take_err() {
+                return Some(Err(DecodeBytesError::InvalidUtf8(e)));
+            }
         }
+        Some(match item? {
+            Ok(b) => Ok(b),
+            Err(e) => Err(DecodeBytesError::DecodeError(e)),
+        })
     }
 
-    fn fold<B, F>(mut self, init: B, mut f: F) -> B
+    fn fold<B, F>(mut self, mut init: B, mut f: F) -> B
     where
         F: FnMut(B, Self::Item) -> B,
     {
-        let b: Result<B, Infallible> = self.0.try_fold(init, |b, item| {
-            Ok(f(b, item.map_err(DecodeBytesError::DecodeError)))
-        });
-        let b = b.unwrap();
-        if let Some(e) = self.0.base_iterator().take_err() {
-            f(b, Err(DecodeBytesError::InvalidUtf8(e)))
-        } else {
-            b
+        loop {
+            let result = self.0.try_fold(init, |b, item| {
+                let item = match item {
+                    Ok(item) => item,
+                    Err(e) => return Err((b, e)),
+                };
+                Ok(f(b, Ok(item)))
+            });
+
+            let (mut b, err) = match result {
+                Ok(b) => (b, None),
+                Err((b, e)) => (b, Some(e)),
+            };
+
+            if let Some(e) = self.0.base_iterator().take_err() {
+                b = f(b, Err(DecodeBytesError::InvalidUtf8(e)));
+            }
+
+            if let Some(e) = err {
+                b = f(b, Err(DecodeBytesError::DecodeError(e)));
+            } else {
+                // Iteration is done because `try_fold` returned `Ok`.
+                return b;
+            }
+            init = b;
         }
     }
 
@@ -610,11 +602,11 @@ where
 
 impl<I: FusedIterator<Item = u8>> FusedIterator for BytesDecoder<I> {}
 
-pub struct StrDecoder<'a>(BaseDecoder<BaseIterator<Chars<'a>>>);
+pub struct StrDecoder<'a>(CharDecoder<Chars<'a>>);
 
 impl<'a> StrDecoder<'a> {
     pub(crate) fn new(s: &'a str, config: DecodeConfig) -> Self {
-        Self(BaseDecoder::new(BaseIterator(s.chars()), config))
+        Self(CharDecoder::new(s.chars(), config))
     }
 }
 
@@ -670,7 +662,7 @@ pub struct DecodeConfig {
     pub require_wrapper: bool,
     /// If true, trailing data after the ending ‘ǲ’ character will be ignored,
     /// rather than causing an error. Additionally, if `require_wrapper` is
-    /// false, extra data before the starting ‘Ǳ’ character will also be
+    /// true, extra data before the starting ‘Ǳ’ character will also be
     /// ignored. [default: false]
     pub relaxed: bool,
 }
